# Autogenerated by Astropy's setup.py on 2017-03-30 22:47:25.701378
from __future__ import unicode_literals
import datetime

version = "1.3.2"
githash = "698ea94934068163e79ee2d632152aabc72611f3"


major = 1
minor = 3
bugfix = 2

release = True
timestamp = datetime.datetime(2017, 3, 30, 22, 47, 25, 701378)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
